/*
    blobby.c
    main program for blobby modeler 1.0
    steven t. huang
    started: 04/14/95
    ended:
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#define LISTMAX          (1000)
#define SQR(x)           ((x)*(x))
#define Tp11             Tp00

typedef struct {
    float                x, y, z, s, zs;
    int                  n, fn;
} listtype;

listtype list[LISTMAX];
int listptr = 0;                         /* allocation index */
float pi = 3.14159265358,
      pi2,                               /* pi/2 */
      Cx, Cy, Cz,                        /* cam position */
      alpha,                             /* x rotation of cam */
      beta,                              /* y rotation of cam */
      theta,                             /* z rotation of cam */
      h = 0.15,                          /* frustrum height (d = 1) */
      Ux, Uy, Uz,                        /* Tview */
      Vx, Vy, Vz,
      Nx, Ny, Nz,
      Xs, Ys, Zs,                        /* screen coordinates */
      Rs,                                /* radius in screen coords */
      Tp00,                              /* Tpersp */
      Tp22 = 1.0,
      Tp23 = 1.0,
      Tp32 = -1.0,
      sina,                              /* cached values */
      cosa,
      sinb,
      cosb,
      sint,
      cost;
int   vtilt = 14,                        /* cam x tilt (* pi/8) */
      htilt = 2,                         /* cam y tilt (* pi/8) */
      rot = 0,                           /* cam rotation */
      freewindow = 1,                    /* free window default on */
      grid = 1;                          /* grid defaults on */
char  dummy[256];

void BMsetup();
void refresh();
void qrefresh();
void drawaxes();

void main(int argc, char **argv)
{
    pi2 = pi/2;
    BMsetup();
    Et_Init(&argc, argv);
    ET_INCLUDE(interface.tcl);
    ET_INSTALL_COMMANDS;
    drawaxes();
    Et_MainLoop();
}

ET_PROC(BMexport)
{
    FILE *f;
    int size, count = 0, x, y, z, i;
    float ox, oy, oz, dx, dy, dz, xx, yy, zz, ss, lx, ly, lz;
    float a = -0.444444, b = 1.888889, c = -2.444444;
    float r2, r4, r6, R2, R4, R6, tmp;

    if (--argc != 8)
        ET(panic "BMexport got %d(argc) parameters expecting 8");
    f = fopen(argv[1], "w");
    if (f == NULL) {
        ET(dialog .exerr {Error!} {Unable to write file} error 0 {Ok});
    }
    else {
        size = atoi(argv[2]);
        ox = atof(argv[3]);
        oy = atof(argv[4]);
        oz = atof(argv[5]);
        dx = atof(argv[6]);
        dy = atof(argv[7]);
        dz = atof(argv[8]);
        fprintf(f, "object 1 class gridpositions counts %d %d %d\n", size,
            size, size);
        fprintf(f, " origin %12f %12f %12f\n", ox, oy, oz);
        fprintf(f, " delta  %12f %12d %12d\n", dx, 0, 0);
        fprintf(f, " delta  %12d %12f %12d\n", 0, dy, 0);
        fprintf(f, " delta  %12d %12d %12f\n", 0, 0, dz);
        fprintf(f, "attribute \"dep\" string \"positions\"\n#\n");
        fprintf(f, "object 2 class gridconnections counts %d %d %d\n", size,
            size, size);
        fprintf(f, "attribute \"element type\" string \"cubes\"\n");
        fprintf(f, "attribute \"ref\" string \"positions\"\n#\n");
        fprintf(f, "object 3 class array type float rank 0 items %d "
            "data follows", size * size * size);
        for (x=0; x<size; x++) {
          xx = 1000.0 * x/(float)size;
          for (y=0; y<size; y++) {
            yy = 1000.0 * y/(float)size;
            for (z=0; z<size; z++) {
              zz = 1000.0 * z/(float)size;
              tmp = 0.0;
              for (i=0; i<listptr; i++) {
                ss = list[i].s*1000.0;
                R2 = ss*ss;
                lx = (list[i].x + 1.0)*500.0;
                ly = (list[i].y + 1.0)*500.0;
                lz = (list[i].z + 1.0)*500.0;
       	        r2 = SQR(xx-lx)+SQR(yy-ly)+SQR(zz-lz);
       	        if (r2 < R2) {
                  R4 = R2 * R2;
                  R6 = R2 * R4;
        	  r4 = r2 * r2;
        	  r6 = r4 * r2;
        	  if (list[i].s > 0)
        	    tmp += (a*r6)/R6+(b*r4)/R4+(c*r2)/R2+1;
        	  else
        	    tmp -= (a*r6)/R6+(b*r4)/R4+(c*r2)/R2+1;
        	}
              }
              if (!count)
                fprintf(f, "\n");
              count = (count+1)%6;
              fprintf(f, "%12f ", tmp);
            }
          }
        }
        fprintf(f, "\nattribute \"dep\" string \"positions\"\n#\n");
        fprintf(f, "object \"default\" class field\n");
        fprintf(f, "component \"positions\" value 1\n");
        fprintf(f, "component \"connections\" value 2\n");
        fprintf(f, "component \"data\" value 3\n");
        fprintf(f, "#\nend\n");
        fclose(f);
    }
    return(ET_OK);
}

ET_PROC(BMadd)
{
    if (listptr < LISTMAX) {
	if (--argc != 5)
            ET(panic "BMadd got only %d(argc) parameters expecting 5");
        list[listptr].x = atof(argv[1]);
        list[listptr].y = atof(argv[2]);
        list[listptr].z = atof(argv[3]);
        list[listptr].s = atof(argv[4]);
        list[listptr].n = atoi(argv[5]);
        listptr++;
    }
    else {
        ET(panic "BMaddlist ran out of space");
    }
    refresh();
    return(ET_OK);
}

ET_PROC(BMdelete)
{
    int i, j, n;
    
    if (--argc != 1)
        ET(panic "BMdelete got %d(argc) parameters expecting 1");
    n = atoi(argv[1]);
    for (i = 0; i < listptr && list[i].n != n; i++);
    if (i == listptr)
        ET(panic "BMdelete can't find node %d(n)");
    else {
        for (j = i+1; j < listptr; j++)
            memcpy(list+j-1, list+j, sizeof(listtype));
        listptr--;
    }
    refresh();
    return(ET_OK);
}

ET_PROC(BMclear)
{
    listptr = 0;
    refresh();
    return(ET_OK);
}

ET_PROC(BMreplace_s)
{
    int i, n;
    
    if (--argc != 2)
        ET(panic "BMreplace_s got %d(argc) parameters expecting 2");
    n = atoi(argv[1]);
    for (i = 0; i < listptr && list[i].n != n; i++);
    if (i == listptr)
        ET(panic "BMreplace_s can't find node %d(n)");
    else
        list[i].s = atof(argv[2]);
    refresh();
    return(ET_OK);
}

ET_PROC(BMreplace_n)
{
    int i, n;
    
    if (--argc != 2)
        ET(panic "BMreplace_n got %d(argc) parameters expecting 2");
    n = atoi(argv[1]);
    for (i = 0; i < listptr && list[i].n != n; i++);
    if (i == listptr)
        ET(panic "BMreplace_n can't find node %d(n)");
    else
        list[i].n = atoi(argv[2]);
    return(ET_OK);
}

ET_PROC(BMtranslate_xy)
{
    int i, n;
    
    if (--argc != 3)
        ET(panic "BMtranslate_xy got %d(argc) parameters expecting 3");
    n = atoi(argv[1]);
    for (i = 0; i < listptr && list[i].n != n; i++);
    if (i == listptr)
        ET(panic "BMtranslate_xy can't find node %d(n)");
    else {
        list[i].x += atof(argv[2]);
        list[i].y -= atof(argv[3]);
    }
    qrefresh(i);
    return(ET_OK);
}

ET_PROC(BMtranslate_yz)
{
    int i, n;
    
    if (--argc != 3)
        ET(panic "BMtranslate_yz got %d(argc) parameters expecting 3");
    n = atoi(argv[1]);
    for (i = 0; i < listptr && list[i].n != n; i++);
    if (i == listptr)
        ET(panic "BMtranslate_yz can't find node %d(n)");
    else {
        list[i].y -= atof(argv[2]);
        list[i].z += atof(argv[3]);
    }
    qrefresh(i);
    return(ET_OK);
}

ET_PROC(BMtranslate_xz)
{
    int i, n;
    
    if (--argc != 3)
        ET(panic "BMtranslate_xz got %d(argc) parameters expecting 3");
    n = atoi(argv[1]);
    for (i = 0; i < listptr && list[i].n != n; i++);
    if (i == listptr)
        ET(panic "BMtranslate_xz can't find node %d(n)");
    else {
        list[i].x += atof(argv[2]);
        list[i].z += atof(argv[3]);
    }
    qrefresh(i);
    return(ET_OK);
}

void alphachg()
{
    alpha = vtilt * pi/8;

    sina = sin(alpha);
    cosa = cos(alpha);

    Cx = 5*cosa*sinb;
    Cy = 5*-sina;
    Cz = 5*cosa*cosb;
    
    /* V = [0 1 0] Rz(theta) Rx(alpha) Ry(beta) */
    Vx = -sint*cosb+cost*sina*sinb;
    Vy = cost*cosa;
    Vz = sint*sinb+cost*sina*cosb;
    
    /* N = [0 0 -1] Rx(alpha) Ry(beta) */
    Nx = -cosa*sinb;
    Ny = sina;
    Nz = -cosa*cosb;

    /* U = N x V */
    Ux = Ny * Vz - Nz * Vy;
    Uy = Nz * Vx - Nx * Vz;
    Uz = Nx * Vy - Ny * Vx;
}

void betachg()
{
    beta = htilt * pi/8;

    sinb = sin(beta);
    cosb = cos(beta);

    Cx = 5*cosa*sinb;
    Cz = 5*cosa*cosb;
    
    /* V = [0 1 0] Rz(theta) Rx(alpha) Ry(beta) */
    Vx = -sint*cosb+cost*sina*sinb;
    Vz = sint*sinb+cost*sina*cosb;
    
    /* N = [0 0 -1] Rx(alpha) Ry(beta) */
    Nx = -cosa*sinb;
    Nz = -cosa*cosb;

    /* U = N x V */
    Ux = Ny * Vz - Nz * Vy;
    Uy = Nz * Vx - Nx * Vz;
    Uz = Nx * Vy - Ny * Vx;
}

void thetachg()
{
    theta = rot * pi/8;

    sint = sin(theta);
    cost = cos(theta);    

    /* V = [0 1 0] Rz(theta) Rx(alpha) Ry(beta) */
    Vx = -sint*cosb+cost*sina*sinb;
    Vy = cost*cosa;
    Vz = sint*sinb+cost*sina*cosb;
    
    /* U = N x V */
    Ux = Ny * Vz - Nz * Vy;
    Uy = Nz * Vx - Nx * Vz;
    Uz = Nx * Vy - Ny * Vx;
}

void BMsetup()
{
    alpha = vtilt * pi/8;
    beta = htilt * pi/8;
    theta = rot * pi/8;
    /* build Tview */
    sina = sin(alpha);
    cosa = cos(alpha);
    sinb = sin(beta);
    cosb = cos(beta);
    sint = sin(theta);
    cost = cos(theta);    

    Cx = 5*cosa*sinb;
    Cy = 5*-sina;
    Cz = 5*cosa*cosb;
    
    /* V = [0 1 0] Rz(theta) Rx(alpha) Ry(beta) */
    Vx = -sint*cosb+cost*sina*sinb;
    Vy = cost*cosa;
    Vz = sint*sinb+cost*sina*cosb;
    
    /* N = [0 0 -1] Rx(alpha) Ry(beta) */
    Nx = -cosa*sinb;
    Ny = sina;
    Nz = -cosa*cosb;

    /* U = N x V */
    Ux = Ny * Vz - Nz * Vy;
    Uy = Nz * Vx - Nx * Vz;
    Uz = Nx * Vy - Ny * Vx;

    /* build Tpersp */
    Tp00 = 1 / h;
}

void toscreen(float Xw, float Yw, float Zw, float s)
{
    float Xv, Yv, Zv;
    
    /* T(-C) */
    Xw -= Cx;
    Yw -= Cy;
    Zw -= Cz;

    /* Tview */
    Xv = Xw * Ux + Yw * Uy + Zw * Uz;
    Yv = Xw * Vx + Yw * Vy + Zw * Vz;
    Zv = Xw * Nx + Yw * Ny + Zw * Nz;

    /* Tpersp */
    Xs = 160 + 100 * ((Xv * Tp00) / Zv);
    Ys = 100 - 100 * ((Yv * Tp11) / Zv);
    Zs = (Zv * Tp22 - Tp32) / Zv;
    
    /* compute radius based on strength */
    Rs = 100 * (s * Tp00) / Zv;
}

void drawaxes()
{
    float x1, y1;
    char *gridc = " -fill SlateGray";
    char *t = " -anchor e -font -adobe-helvetica-bold-r-*-*-12-*-*-*-*-*-*-*";

    /* origin */
    toscreen(0.0, 0.0, 0.0, 0);
    x1 = Xs;
    y1 = Ys;
    
    /* x-axis */
    toscreen(1.0, 0.0, 0.0, 0);
    sprintf(dummy, "%f %f -tags axes", Xs, Ys);
    ET(.fddarea create line %f(x1) %f(y1) %s(dummy) %s(gridc));
    ET(.fddarea create text %s(dummy) -text "x" %s(t));
    
    /* y-axis */
    toscreen(0.0, 1.0, 0.0, 0);
    sprintf(dummy, "%f %f -tags axes", Xs, Ys);
    ET(.fddarea create line %f(x1) %f(y1) %s(dummy) %s(gridc));
    ET(.fddarea create text %s(dummy) -text "y" %s(t));
    
    /* z-axis */
    toscreen(0.0, 0.0, 1.0, 0);
    sprintf(dummy, "%f %f -tags axes", Xs, Ys);
    ET(.fddarea create line %f(x1) %f(y1) %s(dummy) %s(gridc));
    ET(.fddarea create text %s(dummy) -text "z" %s(t));
}

static int fcmp(listtype *i, listtype *j)
{
    if (i->zs > j->zs)
        return(1);
    return(-1);
}

void refresh()
{
    int i;
    float r;

    if (freewindow) {
        ET(.fddarea delete node);
        ET(.fddarea delete axes);
        if (grid)
            drawaxes();
        qsort((char *) list, listptr, sizeof(listtype), (void *) fcmp);
        for (i = 0; i < listptr; i++) {
            toscreen(list[i].x, list[i].y, list[i].z, list[i].s);
            list[i].zs = Zs;
            sprintf(dummy, "freeoval %f %f %f %f", Xs-Rs, Ys-Rs, Xs+Rs, Ys+Rs);
            list[i].fn = ET_INT(%s(dummy));
        }
    }
}

void qrefresh(int i)
{
    if (freewindow) {
	ET(.fddarea delete %d(list[i].fn));
	ET(.fddarea delete %d(list[i].fn+1));
	ET(.fddarea delete %d(list[i].fn+2));
	toscreen(list[i].x, list[i].y, list[i].z, list[i].s);
	sprintf(dummy, "freeoval %f %f %f %f", Xs-Rs, Ys-Rs, Xs+Rs, Ys+Rs);
	list[i].fn = ET_INT(%s(dummy));
    }
}

ET_PROC(BMrefresh)
{
    if (--argc != 2)
        ET(panic "BMrefresh got %d(argc) parameters expecting 2");
    freewindow = atoi(argv[1]);
    grid = atoi(argv[2]);
    refresh();
    return(ET_OK);
}

ET_PROC(BMu)
{
    if (++vtilt > 15)
        vtilt = 0;
    alphachg();
    refresh();
    return(ET_OK);
}

ET_PROC(BMd)
{
    if (--vtilt < 0)
        vtilt = 15;
    alphachg();
    refresh();
    return(ET_OK);
}

ET_PROC(BMr)
{
    if (--htilt < 0)
        htilt = 15;
    betachg();
    refresh();
    return(ET_OK);
}

ET_PROC(BMl)
{
    if (++htilt > 15)
        htilt = 0;
    betachg();
    refresh();
    return(ET_OK);
}

ET_PROC(BMcw)
{
    if (++rot > 15)
        rot = 0;
    thetachg();
    refresh();
    return(ET_OK);
}

ET_PROC(BMccw)
{
    if (--rot < 0)
        rot = 15;
    thetachg();
    refresh();
    return(ET_OK);
}

ET_PROC(BMzo)
{
    h += 0.05;
    if (h > 0.3)
        h = 0.3;
    Tp00 = 1 / h;
    refresh();
    return(ET_OK);
}

ET_PROC(BMzi)
{
    h -= 0.05;
    if (h < 0.05)
        h = 0.05;
    Tp00 = 1 / h;
    refresh();
    return(ET_OK);
}
